/************************************************************************
*
*   VIEWBKUP - Allows you to view a list of the files contained in a
*              DOS 3.3 (or later) BACKUP by reading the CONTROL file.
*              Returns with errorlevel 1 if VIEWBKUP was unsuccessful.
*
*   Method of calling:   VIEWBKUP [d:]
*   
*   Written by Mark Guthrie in Turbo C                         09-25-90
*
************************************************************************/

#include <ctype.h>
#include <dir.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BUFFERSIZE  69
#define MAXDIRNAME  63
#define MAXFILENAME 12

/* convert 2-byte number to unsigned long */
#define SIZE2TOUL(X) ((unsigned long) buffer.uc[X] \
         + ((unsigned long) buffer.uc[X+1]) * 256L)

/* convert 3-byte number to unsigned long */
#define SIZE3TOUL(X) ((unsigned long) buffer.uc[X] \
         + (((unsigned long) buffer.uc[X+1]) * 256L) \
         + (((unsigned long) buffer.uc[X+2]) * 65536L))

enum boolean { FALSE, TRUE };
typedef int boolean;
typedef unsigned long ulong;
typedef unsigned char byte;

/* function prototypes */
int find_control_file(char *fileptr, char drive);
void file_entry(void);
void ultoafmta(ulong ulvalue, char *str);
void display_drive_info(char drive);

/* global variables */
char dirname[MAXDIRNAME+1], olddirname[MAXDIRNAME+1];
char control_file[MAXFILENAME+3];
char drive;
char *monthname[] =
{
    "???","JAN","FEB","MAR","APR","MAY","JUN",
    "JUL","AUG","SEP","OCT","NOV","DEC"
};

FILE *streamptr;    /* pointer to stream */
boolean first_dir;  /* TRUE if first directory */

union BufferUnion
{
    byte uc[BUFFERSIZE+1];      /* unsigned characters */
    char sc[BUFFERSIZE+1];      /* normal characters */
} buffer;

int main(int argc, char *argv[])
{
    if (argc > 1)
        drive = toupper(*(argv[1]));    /* get 1st letter of 1st arg */
    else
        drive = 'A';

    if (find_control_file(control_file,drive) == -1)
    {
        printf("\nControl file not found on drive %c\n",drive);
        exit(1);
    }

    /* open control file for binary read-only & assign pointer to stream */
    if ((streamptr = fopen(control_file,"rb")) == NULL)
    {
        printf("\nError opening %s\n",control_file);
        exit(1);
    }

    /* set streamptr to position 9 */
    fseek(streamptr,9L,SEEK_SET);
    printf("\nContents of Backup disk %3d\n",fgetc(streamptr));

    /* starting with byte 139, read entries until end of file */
    first_dir = TRUE;
    fseek(streamptr,139L,SEEK_SET);
    while (!(feof(streamptr)))
    {
        switch (fgetc(streamptr))
        {
            case 0x22:      /* new file */
                file_entry();
                break;

            case 0x46:      /* new directory */

                /* read directory into buffer */
                fread(buffer.sc,1,BUFFERSIZE,streamptr);

                /* copy dirname to olddirname unless first time thru */
                if (!(first_dir))
                    strcpy(olddirname,dirname);

                /* get dirname from buffer */
                strncpy(dirname,buffer.sc,MAXDIRNAME);
                dirname[MAXDIRNAME] = '\0';

                /* print directory header if dirname <> olddirname
                   or if first time thru */
                if ((first_dir) || (strcmpi(dirname,olddirname) != 0))
                {
                    printf("\nDirectory \\%s\n",buffer.sc);
                    printf("\nFilename Ext        Size   Amt saved"
                           "    Date      Time    Attributes\n");
                }

                first_dir = FALSE;  /* turn off first time thru flag */
                break;
        }
    }
    fclose(streamptr);
    display_drive_info(drive);
    return(0);
}

/****
*
*  Finds the first CONTROL file on the drive specified by first argument.
*  Assumes that fileptr points to a location with at least 15 bytes free.
*  returns with 0 if CONTROL file found, or -1 if no matches are found.
*
****/

int find_control_file (char *fileptr, char drive)
{
    int i;
    struct ffblk fileinfo;

    strcpy(fileptr,"A:\\CONTROL.???");      /* default search string */
    fileptr[0] = drive;                     /* update drive letter */

    /* didn't find match or found match w/o extension */
    if ((findfirst(fileptr,&fileinfo,0) == -1)
      || (!(strpbrk(fileinfo.ff_name,"."))))
        return(-1);

    /* update extension */
    for (i = 11; i <=13; i++)
        fileptr[i] = fileinfo.ff_name[i-3];

    return(0);
}

/****
*
*  Read a file entry from the control file and display
*  the file information.
*
*  Global variables used:
*    buffers, streamptr, monthname
*
*  Record layout of file entry:
*   0 - 11  filename (including period & extension; not null-terminated)
*  13 - 15  file length
*  23 - 25  amount saved on this disk
*       27  file attributes (X X A X X S H R)
*  29 - 30  file time (H H H H H M M M   M M M X X X X X)
*  31 - 32  file date (Y Y Y Y Y Y Y M   M M M D D D D D)
*
****/

void file_entry(void)
{
    char filename[MAXFILENAME+1];
    char charbuffer1[13] = "";
    char charbuffer2[13] = "";
    char *extptr;
    byte fileattr;
    unsigned int month,day,year,hours,minutes;
    ulong filelen,amtsaved,filedate,filetime;
    boolean am;

    /* read file info into buffer */
    fread(buffer.sc,1,33,streamptr);

    /* copy buffer to file name */
    strncpy(filename,buffer.sc,MAXFILENAME);
    filename[MAXFILENAME] = '\0';

    /* replace period with NULL and increment pointer */
    if ((extptr = strchr(filename, '.')) == NULL)
        extptr = &(filename[MAXFILENAME]);      /* no period */
    else
        *extptr++ = '\0';

    /* convert pos 13-15 into unsigned long and format it */
    filelen = SIZE3TOUL(13);
    ultoafmta(filelen,charbuffer1);

    /* convert pos 23-25 into unsigned long and format it */
    amtsaved = SIZE3TOUL(23);
    ultoafmta (amtsaved,charbuffer2);

    /* get filedate in pos 31-32 and convert to day-month-year format */
    filedate = SIZE2TOUL(31);
    if (filedate)
    {
        year = ((filedate & 0xfe00) >> 9) + 80;
        month = ((filedate & 0x01e0) >> 5);
        day = (filedate & 0x001f);
        if (month > 12) month = 0;
    }

    /* get filetime in pos 29-30 and convert to hour:minute format */
    filetime = SIZE2TOUL(29);
    if (filetime)
    {
        hours = (filetime >> 11);
        am = (hours < 12);
        hours = hours % 12;
        if (hours < 1) hours = 12;
        minutes = ((filetime & 0x7e0) >> 5);
    }

    /* get file attributes */
    fileattr = buffer.uc[27];
    
    /* print out filename, extension and sizes */
    printf("%-8s %-3s%12s%12s",filename,extptr,charbuffer1,charbuffer2);

    /* print out date (spaces if filedate is zero) */
    if (filedate)
        printf("  %02d-%s-%02d ",day,monthname[month],year);
    else
        printf("            ");

    /* print out time (spaces if filetime is zero) */
    if (filetime)
    {
        printf(" %2d:%02d",hours,minutes);
        if (am)
            printf("am ");
        else
            printf("pm " );
    }
    else
        printf("       ");

    /* print out attributes (this is clumsy, but it works) */
    if (fileattr & 32)
        printf(" arc");
    else
        printf("    ");
        
    if (fileattr & 1)
        printf(" read");
    else
        printf("     ");
        
    if (fileattr & 2)
        printf(" hid");
    else
        printf("    ");
        
    if (fileattr & 4)
        printf(" sys\n");
    else
        printf("    \n");
}

/****
*
*  Convert an unsigned long integer to a formatted string in buffer
*  The number will be right-justified and will include commas where needed.
*
*  ulvalue (in) : unsigned long integer to convert
*  str (in/out) : pointer to the place to store the formatted string
*
****/

void ultoafmta (ulong ulvalue, char *str)
{
    int i;

    char temp, *savestr = str;

    /*  When loop is done, savestr points to the beginning of
        ulvalue in character form with characters reversed and
        and str points to the character after the last one.
        ex. if ulvalue = 12345 then *savestr = "543,21" and
        str points to the character after the '1'.            */
        
    i = 0;
    do
    {
        i++;
        if (i == 4)
        {
            *str++ = ',';
            i = 1;
        }
        *str++ = (ulvalue % 10) + 48;
        ulvalue /= 10;
    }
    while (ulvalue > 0);

    /* null-terminate *savestr and make str point to the character
       preceding the null character. */
       
    *str-- = '\0';

    /* repeat until savestr and str are equal */
    while (savestr < str)
    {
        temp = *str;
        *str-- = *savestr;
        *savestr++ = temp;
    }
}

/****
*
*  Display total capacity of drive and free space.
*
*  disk (in) : char of drive letter
*
****/

void display_drive_info(char drive)
{
    unsigned long total_space, free_space, bytes_per_cluster;
    struct dfree dfinfo;
    char total_string[15], free_string[15];

    getdfree(drive-64,&dfinfo);
    if (dfinfo.df_sclus == 0xFFFF)
    {
        printf("\nError in getdfree\n");
    }
    else
    {
        bytes_per_cluster = dfinfo.df_sclus * dfinfo.df_bsec;
        total_space = dfinfo.df_total * bytes_per_cluster;
        free_space = dfinfo.df_avail * bytes_per_cluster;
        ultoafmta(total_space,total_string);
        ultoafmta(free_space,free_string);
        printf("\n%s bytes free out of %s total bytes on drive %c\n",
            free_string,total_string,drive);
    }
}
